/* Drawing functions for memory frame buffers.

A frame buffer is an instance of a CMem class and can be shown on
screen by a CScreen instance (on DOS and Linux SvaLib ports). For X it's the window.

Use CDraw to draw in frame buffer: polygons, pixels, ... or clear it.

All files on this directory implement the CDraw class.

FEATURES:
- draw polygons with one color (flat polygons) or with texture. The polygons
  can have any number of vertexes.
- is possible draw some types of polygons making clipping through a edge
  buffer. This edge buffer is used to define the left and right limits of X
  that can be drawn per scanline, providing the clipping. The buffer can
  be defined as the edges of a clipping polygon using the metod
  CDraw::mSetClipEdgeBuffer and can be reseted by CDraw::mResetClipEdgeBuffer
  The clip edge buffer must be an array of type tagInterval with Y resolution
  size. This is the way how portal clipping is implemented.
- draw shaded polygons.

TIPS:
- you can do "enviroment mapping" using a texture as the frame buffer.
- you can save an image in RAW file format saving the frame buffer to a file.
- textures must be with size 256x256.
- the reference system is the same as the video:    (0,0)
                                                      *----->Y
                                                      |
                                                      |
                                                      |
                                                      VX
- for more information about lighting see lighttab.cc

WARNING:
- polygons to be drawn must be convex
- if the library was compiled with lighting/shading;
  to draw textured polygons is needed define a valid light table and
  set CDraw::m_poLightTable to it.
*/



#include <string.h>
#include <math.h>

#include "../defs.h"

#include "../mem.h"

#include "../string.h"

#include "../texture.h"

#include "../lighttab.h"

#include "../asm/draw.inc"

#include "draw.h"



/* constructor - init empty object
**********************************************************/
CDraw::CDraw ()
{
   m_nMidX = m_nMidY = m_nMaxX = m_nMaxY = m_nResX = m_nResY = m_nSize = 0;
   m_nNumLinesAlloc = 0;
   m_poFrBuff = NULL;
   m_pastForwardEdges = m_pastBackwardEdges = NULL;
   m_poLightTable = NULL;
}



/* destructor - empty object
**********************************************************/
CDraw::~CDraw ()
{
   mDestroy ();
}



/* Set object member variables.
oLgt - light table used for polys (can be changed before draw,
       if poly need reflect other color). 
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CDraw::mCreate ()
{
   mDestroy ();
   
   m_poLightTable = NULL;

   return TRUE;
}



/* Empty member variables (memory desalocation)
**********************************************************/
void CDraw::mDestroy ()
{
   // verify if object have some value in variables
   if ( m_poFrBuff != NULL ) {
      m_nMidX = m_nMidY = m_nResX = m_nResY = m_nSize = 0;
      m_poFrBuff = NULL;
      m_poLightTable = NULL;
      m_nNumLinesAlloc = 0;
   }

   // verify if object have some value in variable
   if ( m_poLightTable != NULL )
      m_poLightTable = NULL;

   // verify if edge buffer for forward scan is alocated, if yes desalocate
   if ( m_pastForwardEdges != NULL ) {
      delete m_pastForwardEdges;
      m_pastForwardEdges = NULL;
   }

   // verify if edge buffer for backward scan is alocated, if yes desalocate
   if ( m_pastBackwardEdges != NULL ) {
      delete m_pastBackwardEdges;
      m_pastBackwardEdges = NULL;
   }
}

