/* Gouraud like polys with any number of vertexes. */



#include <string.h>
#include <math.h>

#include "../defs.h"

#include "../mem.h"

#include "../string.h"

#include "../texture.h"

#include "../lighttab.h"

#include "../asm/draw.inc"

#include "draw.h"



/* Draw a interpolated light shaded polygon (gouraud like) with any
number of vertexes (points).
pastXYZes have the array of polygon vertexes with it's light
  intensitys (levels), vertexes in clockwise or anticlockwise order.
payColors have the array of vertexes colors (palette indexes).
Vertexes in clockwise or anticlockwise order.
nNumVertexes is the number of vertexes in array.
Perform frame buffer clip.
**********************************************************/
void CDraw::mIntLgtPolygon ( tagXY* pastXYZes,
                              BYTE* payColors, INT nNumVertexes )
{
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastXYZes [ 0 ].nY,   // y of the top vertex
      nYMax = pastXYZes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastXYZes [ ncVert ].nY > nYMax ) {
         nYMax = pastXYZes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastXYZes [ ncVert ].nY < nYMin ) {
         nYMin = pastXYZes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort polygon fill
   if ( nYMax == nYMin )
      return;

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanIntLgtEdge ( m_pastForwardEdges,
                         pastXYZes [ nPos1 ], pastXYZes [ nPos2 ],
                         payColors [ nPos1 ], payColors [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanIntLgtEdge ( m_pastBackwardEdges,
                         pastXYZes [ nPos1 ], pastXYZes [ nPos2 ],
                         payColors [ nPos1 ], payColors [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // draw polygon
   // define scanline Y (if nYMin is less than 0, outside frame buff, use 0)
   INT
      nY = nYMin < 0 ? 0 : nYMin;
   // define memory position
   PTRVAL
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + nY * m_nResX;

   // if X values for this scanline are swapped
   if ( m_pastBackwardEdges [ nY ].nX < m_pastForwardEdges [ nY ].nX )

      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
   
         mHorizontalIntLgtLine ( m_pastBackwardEdges [ nY ],
                                  m_pastForwardEdges [ nY ],
                                  pLineAddress );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
   else
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
   
   
         mHorizontalIntLgtLine ( m_pastForwardEdges [ nY ],
                                  m_pastBackwardEdges [ nY ],
                                  pLineAddress );
         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
}



/* Scan edges.
pastEdge - fill the x coordenates for each scanline 
and the light for each scanline.
stV1 and stV2 - vertex between scan must be done.
yColor1 and yColor2 - the vertexes colors (palette indexes)
Always stV1 is less (topper) than stV2 in Y axis.
**********************************************************/
void CDraw::mScanIntLgtEdge ( tagPolyEdge* pastEdge,
                               tagXY& stV1, tagXY& stV2,
                               BYTE yColor1, BYTE yColor2 )
{
   // verify if Y are in same line, if yes exit
   if ( stV1.nY == stV2.nY )
      return;

   INT
      // define where X start in first scanline, use fix point 16.16
      nX = stV1.nX << 16,
      // X increase rate
      nXInc = ( ( stV2.nX - stV1.nX ) << 16 ) /
                ( stV2.nY - stV1.nY );

   // do Y clip in frame buffer limits
   INT
      nYTop,
      nYBottom;
   // if y for vertex 1 is upper than frame buffer limits
   if ( stV1.nY < 0 ) {
      nYTop = 0;
      // if top Y has been cliped then is necessary update initial X value
      nX -= stV1.nY * nXInc;
   }
   else
      nYTop = stV1.nY;
   // if y for vertex 2 is lower than frame buffer limits
   if ( stV2.nY > m_nMaxY )
      nYBottom = m_nMaxY;
   else
      nYBottom = stV2.nY;

   // verify if need clip
   if ( nYBottom <= nYTop )
      return;

   // define the color parameters
   INT
      nColor = ( (INT) yColor1 ) << 16,                        // initial color
      nColorInc = ( ( (INT) yColor2 - yColor1 ) << 16 ) /      // increase rate
                    ( stV2.nY - stV1.nY );

   // define X values in each scanline between top vertex and bottom vertex
   INT
      nYScan = nYTop;    // scanline in test
   do {

      pastEdge [ nYScan ].nX = nX >> 16; // X in frame buff, convert fix->int
      pastEdge [ nYScan ].nColor = nColor;   // color
      nX += nXInc;                       // new X in next scanline
      nColor += nColorInc;               // new color in next scanline
      nYScan ++;                         // next scanline

   } while ( nYScan <= nYBottom );       // while not found bottom vertex
}



/* Draw a horizontal gouraud shaded like line.
Clip X values through given clip edge X limits.
stEdgeIni - left edge values (X and texture start)
stEdgeEnd - right edge values (X and texture end)
pLineAddress is the position in frame buffer where the
line must be drawn.
**********************************************************/
void CDraw::mHorizontalIntLgtLine ( tagPolyEdge& stEdgeIni,
                                     tagPolyEdge& stEdgeEnd,
                                     PTRVAL pLineAddress )
{
   // clip x values
   INT
      nFrXIni = stEdgeIni.nX,   // define left X value in frame buffer
      nFrXEnd = stEdgeEnd.nX;   // define right X value in frame buffer
   // verify start
   if ( stEdgeIni.nX < 0 )
      nFrXIni = 0;
   // verify end
   if ( stEdgeEnd.nX > m_nMaxX )
      nFrXEnd = m_nMaxX;

   // verify if line has no lenght
   if ( nFrXEnd <= nFrXIni )
      return;

   INT
      // define start color
      nColor = stEdgeIni.nColor,
      // color increase rate
      nColorInc = ( stEdgeEnd.nColor - stEdgeIni.nColor ) / ( stEdgeEnd.nX - stEdgeIni.nX );

   // if X has been clipped
   if ( stEdgeIni.nX < 0 )
      nColor -= stEdgeIni.nX * nColorInc;

   // define start and end address in frame buffer
   BYTE
      *pyCurr = (BYTE*) pLineAddress + nFrXIni,   // current address
      *pyEnd = (BYTE*) pLineAddress + nFrXEnd;    // end address

   // line draw loop
   do {

      // set the pixel
     *pyCurr = nColor >> 16;

      // update color
      nColor += nColorInc;
      // go to next pixel
      pyCurr ++;

   } while ( pyCurr <= pyEnd );  // while not found end of line
}

