/* Graphic/aplication management for X Window System.

Implement a X window app instance with C++ and event handlers with C.

Possible:
create one app window, init app, set palette, close window,
update the window, set the event loop

TIPS:
- the reference system is the same as the video:   (0,0)
                                                      *----->Y
                                                      |
                                                      |
                                                      |
                                                      VX
- to create one X Window app follow the steps: create the app with mCreate, set the
  event loop with mExecApp; the app exit by your self (without external function call).
- the macro COMP_SHAREMEMORY - define the use of shared memory, speeding up the output.
  (see config.h for more information).
- Palette emulation - if video is using more than 8BPP this class emulate a palette mode.
  Because 3DFC beta 8 only works at 8BPP, it only draw on an 8BPP frame buffer. Under
  emulation mode this class create one virtual palette and give for the engine one
  8BPP frame buffer. Before show the frame it is copyed for one nBPP drame buffer
  using the virtual palette RGBs as RGB source.
  
WARNING:
- the X type KeySym must be castable to DWORD
*/



#include <time.h> 
#include <sys/timeb.h>

// need to this class:
#include "../../defs.h"
#include "../../mem.h"

// to include engine:
#include "../../string.h"

#include "../../texture.h"

#include "../../file.h"

#include "../../filerem.h"

#include "../../filepcx.h"

#include "../../lighttab.h"

#include "../../draw/draw.h"

#include "../../node.h"

#include "../../list.h"

#include "../../polygon.h"

#include "../../nodepol.h"

#include "../../socket.h"

#include "../../svsocket.h"

#include "../../nodexyz.h"

#include "../../map.h"

#include "../../math.h"

#include "../../mesh.h"

#include "../../sector.h"

#include "../../nodesec.h"

#include "../../refsys.h"

#include "../../camsys.h"

#include "../../object.h"

#include "../../nodeint.h"

#include "../../nodetex.h"

#include "../../nodeobj.h"

#include "../../nodexyz.h"

#include "../../eventtab.h"

#include "../../comp/nodecomp.h"

#include "../../comp/cokbcont.h"

#include "../../comp/codata.h"

#include "../../comp/cocamsys.h"

#include "../../comp/comesh.h"

#include "../../filetxe.h"

#include "../../scenery.h"

#include "../../render/render.h"

#include "../../engine.h"

#include "keyboard.h"

// need to this class:
#include "xwinapp.h"



// Declare a X Window app instance
CXWindowApp
   oAPP;



// data to the X system
static XrmOptionDescRec astTABLE [] = {
   { "-flag",   "Flag",                XrmoptionNoArg,  (caddr_t) "on" },
   { "-param",  "parameter",           XrmoptionSepArg, NULL },
};

typedef struct {
   INT nFlag;
   INT nParameter;
   char *pyGeometry;
   INT nBorderwidth;
} pAppResourceRec, *pAppRes;

static pAppResourceRec stAPP_RESOURCES;
static char szWIN_DIMENSION [ 100 ];

static XtResource astRESOURCES [] = {
  { "param", "Parameter", XtRInt, sizeof ( INT ),
    XtOffset ( pAppRes, nParameter ), XtRImmediate, (caddr_t) 80},
  { "flag", "Flag", XtRBoolean, sizeof ( Boolean ),
    XtOffset ( pAppRes, nFlag ), XtRImmediate, (caddr_t) TRUE },
  { XtNgeometry,  "Geometry", XtRString, sizeof ( caddr_t ),
    XtOffset ( pAppRes, pyGeometry ), XtRString, (caddr_t) szWIN_DIMENSION },
  { XtNborderWidth, "borderWidth", XtRInt, sizeof ( INT ),
    XtOffset ( pAppRes, nBorderwidth ), XtRImmediate, (caddr_t) 1},
};



// the message handlers
void WindowHandler (  Widget stWi, XtPointer stClosure, XEvent* pstEvent,
                      Boolean* bCont_to_dispatch );
void EngineHandler ( XtPointer, XtIntervalId * );



/* constructor
**********************************************************/
CXWindowApp::CXWindowApp ()
{
   m_pstDisplay = NULL;
   m_stTopLevel = 0;
   m_stShell = 0;
   m_stBox = 0;

   // window data
   m_nScreenNum = 0;
   m_nDpyWidth = 0;
   m_nDpyHeight = 0;
   m_pXImg = NULL;
   m_pyVidPal = m_pyVid = NULL;

   m_nBpp = 0;

   m_stColorMap = 0;
   szWIN_DIMENSION [ 0 ] = 0;
   m_dwResX = m_dwResY = 0;

   m_poEngine = NULL;
}



/* destructor
**********************************************************/
CXWindowApp::~CXWindowApp ()
{
}



/* Create a window, init event handler,...
**********************************************************/
void CXWindowApp::mCreate ( char* szWinTitle, DWORD dwResX, DWORD dwResY,
                            INT nArgc, char* pszArgv [] )
{
   // define the resolution
   m_dwResX = dwResX;
   m_dwResY = dwResY;
   sprintf ( szWIN_DIMENSION, "%dx%d", m_dwResX, m_dwResY );

   // init the app and window

   m_stTopLevel = XtVaAppInitialize ( &m_stApContext, szWinTitle,   
                  astTABLE, XtNumber ( astTABLE ), &nArgc, pszArgv,
                  NULL, NULL, 0 );

   XtGetApplicationResources ( m_stTopLevel, &stAPP_RESOURCES, astRESOURCES,
                               XtNumber ( astRESOURCES ), NULL, (Cardinal) 0 );

   //XParseGeometry (stAPP_RESOURCES.geometry, &geom_x, &geom_y, &geom_width,
   //&geom_height);

   m_stShell = XtVaCreatePopupShell ( szWinTitle,
               topLevelShellWidgetClass,
               m_stTopLevel,
               XtNgeometry, stAPP_RESOURCES.pyGeometry,
               XtNborderWidth, stAPP_RESOURCES.nBorderwidth,
               0 );

   //stBox = XtCreateManagedWidget ("box", boxWidgetClass, stShell, NULL, (Cardinal) 0);
   m_stBox = XtCreateWidget ( "box", boxWidgetClass, m_stShell, NULL, (Cardinal) 0 );

   XtPopup ( m_stShell, XtGrabNone );
 
   // get some new window data
   Screen*
      pScreenPtr;
   m_pstDisplay = XtDisplay ( m_stShell );
   m_nScreenNum = DefaultScreen ( m_pstDisplay );
   pScreenPtr   = DefaultScreenOfDisplay ( m_pstDisplay );
   m_nDpyWidth  = DisplayWidth ( m_pstDisplay, m_nScreenNum );
   m_nDpyHeight = DisplayHeight ( m_pstDisplay, m_nScreenNum );
   m_nBpp       = DefaultDepth ( m_pstDisplay, m_nScreenNum );
   m_stWindow   = XtWindow ( m_stShell );

   // configure drawing attributes
   XGCValues
      stValues;
   m_stGC = XCreateGC ( m_pstDisplay, m_stWindow, 0, &stValues );
   XSetForeground ( m_pstDisplay, m_stGC, BlackPixel ( m_pstDisplay, m_nScreenNum ) );
   XSetLineAttributes ( m_pstDisplay, m_stGC, 0, LineSolid, CapButt, JoinMiter );
   XSetGraphicsExposures ( m_pstDisplay, m_stGC, False );

   // create palette
   if ( m_nBpp == 8 )
      m_stColorMap = XCreateColormap ( m_pstDisplay, RootWindow ( m_pstDisplay, m_nScreenNum ), DefaultVisual ( m_pstDisplay, m_nScreenNum ), AllocAll );

   // set window message handler
   XtAddEventHandler ( m_stShell,
                       ExposureMask | KeyPressMask | KeyReleaseMask |
                       KeymapStateMask | FocusChangeMask |
                       PointerMotionMask | ButtonPressMask | ButtonReleaseMask, True,
                       WindowHandler, NULL );

   XMapWindow ( m_pstDisplay, XtWindow ( m_stShell ) );

   // set palette
   if ( m_nBpp == 8 )
      XSetWindowColormap ( m_pstDisplay, XtWindow ( m_stShell ), m_stColorMap );

   // init frame rate counter
   m_stClockStart = clock ();
   m_dwFrames = 0;
}



/* Close the window, quit the app
**********************************************************/
void CXWindowApp::mDestroy ()
{
   if ( m_stShell )
     XtPopdown ( m_stShell );
   if ( m_stTopLevel )
     XtDestroyWidget ( m_stTopLevel );

   m_poEngine = NULL;
   m_pyVid = NULL;
   if ( m_nBpp != 8 ) {     // desalocate indexed engine frame buffer
      delete m_pyVidPal;
      m_pyVidPal = NULL;
   }   

   // enable autorepeat
   if ( m_pstDisplay )
      XAutoRepeatOn ( m_pstDisplay );

   // end clock count and show message
   clock_t
      stClockEnd = clock ();
   float
      fDeltaTime = (float) stClockEnd - m_stClockStart;
   printf ( MSG_FRAMES, fDeltaTime / CLOCKS_PER_SEC, oAPP.m_dwFrames, (float) oAPP.m_dwFrames / ( fDeltaTime / CLOCKS_PER_SEC ) );
   printf ( "(timming maybe wrong)" );

   //exit ( 0 );

   printf ( "\n\nHey!   Press Control-C to end exit!!!\n" );
}



/* Enter in the app loop.
oEngine - the engine
**********************************************************/
void CXWindowApp::mExecApp ( CEngine& oEngine )
{
   m_poEngine  = &oEngine;

   XtAppAddTimeOut ( m_stApContext, 300, EngineHandler, NULL );
   XtAppMainLoop ( m_stApContext );
}



/* Write a memory array of RGB data in palette.
**********************************************************/
void CXWindowApp::mWritePal ( CMem& oMem )
{
   // seek all palette writing RGB
   for ( DWORD dwcCol = 0; dwcCol < 768; dwcCol += 3 ) {

      // define RGB

      // ...* 4 to make palette more lighted (the scenery RGB range is 0..63)
      // ...* 256 to convert for X convensions (16 bits?)
      DWORD
         dwRed   = oMem.m_payMem [ dwcCol + 0 ] * 4,
         dwGreen = oMem.m_payMem [ dwcCol + 1 ] * 4,
         dwBlue  = oMem.m_payMem [ dwcCol + 2 ] * 4;
         
      // set RGB

      // if running at 8BPP, set X palette
      if ( m_nBpp == 8 ) {
 
         XColor
            stColor;
         stColor.pixel = dwcCol / 3;
         stColor.red   = dwRed   * 256;
         stColor.green = dwGreen * 256;
         stColor.blue  = dwBlue  * 256;
         stColor.flags = DoRed | DoGreen | DoBlue;
         XStoreColor ( m_pstDisplay, m_stColorMap, &stColor );
      }
      else {       // set emulation palette (more than 8BPP)
      
         tagRGB
            *pstPalEmul = m_astPalEmul + dwcCol / 3;
         pstPalEmul->dwR = dwRed;
         pstPalEmul->dwG = dwGreen;
         pstPalEmul->dwB = dwBlue;
      }   
   }
}



/* Show a image buffer on window area.
If emulating 8BPP mode, make conversion from 8BPP -> nBPP;
n is current X Window bits per pixel.
**********************************************************/
void CXWindowApp::mShow ()
{
   // if emulating palette mode (X deph is more than 8),
   // need translate the current engine frame buffer to image frame buffer
   if ( m_nBpp != 8 ) {

      // current pixel to convert
      BYTE
         *pyPixelPal =  m_pyVidPal,
         *pyEndPal = m_pyVidPal + m_dwResX * m_dwResY - 1,
         *pyPixelVid = m_pyVid;
         
      // scan engine frame buffer   
      while ( pyPixelPal != pyEndPal ) {
         
         // set X Window image
         switch ( m_nBpp ) {
         
            case 15:
            case 16:
               *( (UINT16*) pyPixelVid ) = (UINT16) mFormatPixelColor ( *pyPixelPal );
               pyPixelVid += 2;
               break;
         
            case 32:
               *( (DWORD*) pyPixelVid ) = mFormatPixelColor ( *pyPixelPal );
               pyPixelVid += 4;
               break;
         
            case 24:
               *( (DWORD*) pyPixelVid ) = mFormatPixelColor ( *pyPixelPal );
               pyPixelVid += 3;
               break;
         }     
         
         pyPixelPal ++;
      }   
   }   
   
   #ifdef COMP_SHAREMEMORY
   XShmPutImage ( m_pstDisplay, m_stWindow, m_stGC, &m_stShmImg, 0, 0, 0, 0, m_dwResX, m_dwResY, False );
   #else
   XPutImage ( m_pstDisplay, m_stWindow, m_stGC, m_pXImg, 0, 0, 0, 0, m_dwResX, m_dwResY );
   #endif
}



/* Receive one palette color and return the RGB color, using
the X Window BPP.
At 8BPP just return the given index color.
dwPalCol - the indexed palette color
return the formated color
**********************************************************/
DWORD CXWindowApp::mFormatPixelColor ( DWORD dwPalCol )
{
   // if X Window at 8BPP
   if ( m_nBpp == 8 )
      return dwPalCol;
      
   // if W bpp is more than 8   

   // set color acording current X Window BPP
   tagRGB
      *pstPal = m_astPalEmul + dwPalCol;   // point the RGB for the indexed color

   // return the RGB on the correct format
   switch ( m_nBpp ) {

      case 16:             // 16 bpp
         {
            tagFormat16
               unRet;
            // clear
            unRet.nValue = 0;   
            // set 
            unRet.stBits.bitsR = ( pstPal->dwR & 0xF8 ) >> 3;
            unRet.stBits.bitsG = ( pstPal->dwG & 0xFC ) >> 2;
            unRet.stBits.bitsB = ( pstPal->dwB & 0xF8 ) >> 3;
            // return
            return unRet.nValue;
         }   
         break;

      case 15:             // 15 bpp
         {
            tagFormat15
               unRet;
            // clear
            unRet.nValue = 0;   
            // set 
            unRet.stBits.bitsR = ( pstPal->dwR & 0xF8 ) >> 3;
            unRet.stBits.bitsG = ( pstPal->dwG & 0xF8 ) >> 3;
            unRet.stBits.bitsB = ( pstPal->dwB & 0xF8 ) >> 3;
            // return
            return unRet.nValue;
         }   
         break;
 
      default:            // 24, 32 bpp
         return pstPal->dwB + ( pstPal->dwR << 8 ) + ( pstPal->dwB << 16 );
         break;
   }
 
   return 0;
}



/* This is the window event handler - receive input from the window
**********************************************************/
void WindowHandler (  Widget stWi, XtPointer stClosure, XEvent* pstEvent,
                      Boolean* bCont_to_dispatch )
{
   // disable GCC compiling warnings
   (void)stWi; (void)stClosure; (void)bCont_to_dispatch;

   switch ( pstEvent->type ) {

      case Expose:
         
         // set output buffer
         if ( ! oAPP.m_pXImg ) {

            oAPP.m_pXImg = XGetImage ( oAPP.m_pstDisplay, oAPP.m_stWindow, 0, 0, oAPP.m_dwResX, oAPP.m_dwResY, AllPlanes, ZPixmap );

            #ifdef COMP_SHAREMEMORY

            oAPP.m_stShmImg = *oAPP.m_pXImg;
            oAPP.m_stShmi.shmid = shmget ( IPC_PRIVATE, oAPP.m_pXImg->bytes_per_line * oAPP.m_pXImg->height, IPC_CREAT | 0777 );
            oAPP.m_stShmi.shmaddr = (char*) shmat ( oAPP.m_stShmi.shmid, 0, 0 );
            oAPP.m_stShmi.readOnly = FALSE;
            XShmAttach ( oAPP.m_pstDisplay, &oAPP.m_stShmi );
            oAPP.m_stShmImg.data = oAPP.m_stShmi.shmaddr;
            oAPP.m_pyVid = (unsigned char*) oAPP.m_stShmi.shmaddr;
            oAPP.m_stShmImg.obdata = (char*) &oAPP.m_stShmi;

            #else

            oAPP.m_pyVid = (unsigned char*) ( oAPP.m_pXImg->data );

            #endif
         }
         
         // set engine frame buffer
         if ( oAPP.m_nBpp != 8 )
            // if emulating 8BPP, alloc the indexed frame buffer
            oAPP.m_pyVidPal = new BYTE [ oAPP.m_dwResX * oAPP.m_dwResY ];
         else
            // if at 8BPP, no emulation need - indexed frame buffer is the image buffer
            oAPP.m_pyVidPal = oAPP.m_pyVid;   

         // update engine frame buffer
         if ( oAPP.m_poEngine != NULL ) {

            oAPP.m_oFrameBufferRef.mSetBuffer ( (void*) oAPP.m_pyVidPal, oAPP.m_dwResX * oAPP.m_dwResY, FALSE );
            oAPP.m_poEngine->mSetFrameBuffer ( oAPP.m_dwResX, oAPP.m_dwResY, oAPP.m_oFrameBufferRef );
         }
         break;

      case FocusIn:
         // disable keyboard autorepeat (to avoid multiple kb msgs for same key press)
         if ( oAPP.m_pstDisplay )
            XAutoRepeatOff ( oAPP.m_pstDisplay );
         break;

      case FocusOut:
         if ( oAPP.m_pstDisplay )
            XAutoRepeatOn ( oAPP.m_pstDisplay );

         break;

      case KeyPress:
         {
            KeySym
               stKey;
            int
               nKey2,
               nState;
 
            stKey = XLookupKeysym ( (XKeyEvent*) pstEvent, 0 );
            nState = ( (XKeyEvent*) pstEvent )->state;
            nKey2 = stKey;
            
            // set the key in key table
            oAPP.m_oKb.mSetKeyTable ( stKey, TRUE );
            
            //printf("key press: %d    %d %d\n",stKey, nKey2, nState);
 
            // X specific action
            //switch ( stKey ) {
            //}
         }
         break;
 
      case KeyRelease:
         {
 
            KeySym
               stKey;
            //int
              // nKey2,
              // nState;

            stKey = XLookupKeysym ( (XKeyEvent*) pstEvent, 0 );
            
            // reset the key in key table
            oAPP.m_oKb.mSetKeyTable ( stKey, FALSE );
 
            // X specific action
            switch ( stKey ) {
 
               case XK_Escape:
                   oAPP.mDestroy ();
                   break;

               default:
                  break;
            }
         }
         break;

      default:
         break;
   }
}



/* This is the engine handler - run one engine loop interaction.
**********************************************************/
void EngineHandler ( XtPointer, XtIntervalId * )
{
   // verify if exist frame buffer (it exit when window message Expose is received)
   if ( oAPP.m_pyVid && oAPP.m_poEngine ) {

      oAPP.m_poEngine->mRun ();
      oAPP.mShow ();

      oAPP.m_dwFrames ++;
   }

   XtAppAddTimeOut ( oAPP.m_stApContext, 1, EngineHandler, NULL );
}

